#include <errno.h>
#include <grp.h>
#include <pwd.h>
#include <stdio.h>
#include <string.h>
#include <sys/acl.h>
#include <sys/stat.h>

#define CHUNK 1024
int acl_output_order[] =  { ACL_USER_OBJ,  ACL_USER, ACL_GROUP_OBJ, 
			      ACL_GROUP, ACL_MASK, ACL_OTHER_OBJ };

static int append(char**buffer, int *size, const char *text)
{
	char *tmp;
	int n1 = 0;
	int n2 = strlen(text);
	if(*buffer)
	  n1 = strlen(*buffer);
	if(n1 + n2 + 1 > *size) {
		*size = n1 + n2 + 1 + CHUNK;
		tmp = (char*)realloc(*buffer, *size);
		if(!tmp) {
			if(*buffer)
			  free(*buffer);
			*buffer = NULL;
			*size = 0;
			return 0;
		} else {
			if(!(*buffer))
			  *tmp = '\0';
			*buffer = tmp;
		}
	}
	strcat(*buffer, text);
	return 1;
}

void
perm2string(int perms, char *perm)
{
	perm[0] = (perms & ACL_READ) ? 'r' : '-';
	perm[1] = (perms & ACL_WRITE) ? 'w' : '-';
	perm[2] = (perms & ACL_EXECUTE) ? 'x' : '-';
	perm[3] = 0;
}

static
char  *_acl_to_text (const acl_t acl_d, ssize_t *len_buffer)
{
	char  *buffer = NULL;
	int  size = 0;
	char  perm[4];
	char  id[32];
	int j, ret = 1;
	acl_entry_t acle;
	acl_tag_t tag_type;
	struct passwd *pw;
	struct group *gr;
	int mask = 7;		/* used to mask ACL_USER,ACL_GROUP_OBJ,ACL_GROUP entries */
	int got_mask = 0, calc_mask = 0; /* calculate mask if no ACL_MASK present */

	CHECK_ACL_VALID (acl_d, NULL);
	append(&buffer, &size, "");
	for (j = 0; j < NO_ACL_TYPES; j++) {
		tag_type = acl_output_order[j];
		for (acle = acl_d->acl_faclep; acle; acle = acle->acle_next) {
			if (acle->acle_tag_type == ACL_MASK)
			  mask = acle->acle_perms;
			if (acle->acle_tag_type != tag_type)
			  continue;
			switch (acle->acle_tag_type) {
			      case ACL_USER_OBJ:
				ret = ret && append(&buffer, &size, "user:");
				break;
			      case ACL_GROUP_OBJ:
				ret = ret && append(&buffer, &size, "group:");
				calc_mask |= acle->acle_perms;
				break;
			      case ACL_OTHER_OBJ:
				/*if(!got_mask){
				  ret = ret && append(&buffer, &size, "mask::");
				  perm2string(calc_mask, perm);
				  ret = ret && append(&buffer, &size, perm);
				  ret = ret && append(&buffer, &size, "\n");
				  }*/
				ret = ret && append(&buffer, &size, "other:");
				break;
			      case ACL_MASK:
				ret = ret && append(&buffer, &size, "mask:");
				got_mask = 1;
				break;
			      case ACL_USER:
				ret = ret && append(&buffer, &size, "user:");
				calc_mask |= acle->acle_perms;
				pw = getpwuid ((uid_t)acle->acle_tag);
				if (!pw) {
					sprintf (id, "%d", (uid_t)acle->acle_tag);
					ret = ret && append(&buffer, &size, id);
				} else {
					ret = ret && append(&buffer, &size, pw->pw_name);
				}
				break;
			      case ACL_GROUP:
				ret = ret && append(&buffer, &size, "group:");
				calc_mask |= acle->acle_perms;
				gr = getgrgid ((gid_t)acle->acle_tag);
				if (!gr) {
					sprintf (id, "%d", (gid_t)acle->acle_tag);
					ret = ret && append(&buffer, &size, id);
				} else {
					ret = ret && append(&buffer, &size, gr->gr_name);
				}
				break;
			      default:
				errno = EINVAL;
				if(buffer)
				  free (buffer);
				return NULL;
			}
			ret = ret && append(&buffer, &size, ":");
			perm2string(acle->acle_perms, perm);
			ret = ret && append(&buffer, &size, perm);
			if((acle->acle_perms & mask) != acle->acle_perms) {
				ret = ret && append(&buffer, &size, " #effective: ");
				perm2string(acle->acle_perms & mask, perm);
				ret = ret && append(&buffer, &size, perm);
			}
			ret = ret && append(&buffer, &size, "\n");
		}
	}
	if (len_buffer){
		if (buffer){
			*len_buffer = strlen (buffer) + 1;
		} else {
			*len_buffer = 0;
		}
	}
	return buffer;
}

char  *acl_to_text (const acl_t acl_d, ssize_t *len_buffer)
{
	return _acl_to_text (acl_d, len_buffer);
}

char  *acl_to_long_text (const acl_t acl_d, ssize_t *len_buffer)
{
	return _acl_to_text (acl_d, len_buffer);
}
