#include <grp.h>
#include <pwd.h>
#include <stdio.h>
#include <sys/acl.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>

#include "util.h"

extern const char *__progname;

static const char *program_name;

static
volatile void	usage (void)
{
	fprintf (stderr, "Usage: %s [-d] [file...]\n", program_name);
	exit (1);
}

static
int getfacl (const char *pathname, acl_type_t type)
{
	acl_t		acl;
	char		*text_acl;
	ssize_t		len;
	int		ret = 0;
	struct passwd	*pw;
	struct group	*gr;
	struct stat	st;

	acl = acl_get_file (pathname, type);
	if (!acl && errno == EINVAL) {
		acl = acl_from_mode_file(pathname, type);
	}
	if (!acl) {
		perror (pathname);
		ret = 2;
	} else {
		text_acl = acl_to_long_text (acl, &len);
		if (!text_acl) {
			perror ("acl_to_text");
			ret = 4;
		} else {
			if (stat (pathname, &st) == -1) {
				perror ("stat");
				ret = 8;
			} else {
				printf ("# file:%s\n", pathname);
				pw = getpwuid (st.st_uid);
				/* POSIX: Only numerical IDs. */
				if (pw)
				  printf ("# owner:%s\n",
					  pw->pw_name);
				else
				  printf ("# owner:%d\n",
					  st.st_uid);
				gr = getgrgid (st.st_gid);
				if (gr)
				  printf ("# group:%s\n",
					  gr->gr_name);
				else
				  printf ("# group:%d\n",
					  st.st_gid);
				printf ("%s", text_acl);
			}
		}
		acl_free_text (text_acl);
	}
	return ret;
}

static
int		getfacl_from_stdin (acl_type_t type)
{
	char		file[MAXPATHLEN];
	int		ret = 0;
	int first = 1;
	while (fgets (file, sizeof (file), stdin) != NULL) {
		if(!first)
		  printf("\n");
		first = 0;
		ret |= getfacl (strtok(file,"\n"), type);
	}
	return ret;
}

int
main (int argc, char *argv[])
{
	acl_type_t	type = ACL_TYPE_ACCESS;
	int		i;
	char		c;
	int		ret = 0;

	program_name = __progname;
	while ((c = getopt (argc, argv, "d")) != EOF)
	  switch (c)
{
      case 'd':
	type = ACL_TYPE_DEFAULT;
	break;
      default:
	usage ();
}

	if (optind == argc)
	  ret |= getfacl_from_stdin (type);
	else {
		for (i = optind; i < argc; i++) {
			if(i > optind)
			  printf("\n");
			if (strcmp (argv[i], "-") == 0)
			  ret |= getfacl_from_stdin (type);
			else
			  ret |= getfacl (argv[i], type);
		}
	}
	exit (ret);
}
