#include <errno.h>
#include <stdio.h>
#include <sys/acl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>

void perror_with_acl (const acl_t acl, const char *s)
{
	char		*acl_text;
	ssize_t		len;
	int		err;

	err = errno;
	acl_text = acl_to_text (acl, &len);
	if (acl_text) {
		fprintf (stderr, "%s (%s): %s\n", s, acl_text, strerror (err));
		acl_free_text (acl_text);
	} else
	  fprintf (stderr, "%s (acl_to_text failed!): %s\n", s,
		   strerror (err));
}

void perror_with_arg (const char *s1, const char *s2)
{
	fprintf (stderr, "%s: %s: %s\n", s1, s2, strerror (errno));
}

void perror_with_acl_and_arg (const acl_t acl, const char *s1, const char *s2)
{
	char		*acl_text;
	ssize_t		len;
	int		err;

	err = errno;
	acl_text = acl_to_text (acl, &len);
	if (acl_text) {
		fprintf (stderr, "%s (%s): %s: %s\n", s1, acl_text, s2,
			 strerror (err));
		acl_free_text (acl_text);
	} else
	  fprintf (stderr, "%s (acl_to_text failed!): %s: %s\n", s1, s2,
		   strerror (err));
}

void print_acl (FILE *stream, const acl_t acl)
{
	char		*acl_text;
	ssize_t		len;

	acl_text = acl_to_text (acl, &len);
	if (acl_text) {
		fprintf (stream, "%s", acl_text);
		acl_free_text (acl_text);
	} else
	  fprintf (stream, "(acl_to_text failed!)");
}

acl_t
acl_from_mode_file (const char *pathname, acl_type_t type)
{
	acl_t		acl;
	acl_entry_t	acle;
	struct stat buf;

	if(stat(pathname, &buf) != 0)
	  return NULL;

	acl = acl_init (sizeof (struct acl));
	if (!acl)
	  return NULL;

	acle = NULL;
	if (acl_create_entry (&acl, &acle) != 0)
	  goto error;
	if (acl_set_tag_type (acle, ACL_USER_OBJ) != 0)
	  goto error;
	if (acl_set_mode (acle, ACL_ABSOLUTE_ENTRY) != 0)
	  goto error;
	acl_clear_perms(&acle->acle_perms);
	if(buf.st_mode & S_IRUSR)
	  if (acl_add_perm (&acle->acle_perms, ACL_READ) != 0)
	    goto error;
	if(buf.st_mode & S_IWUSR)
	  if (acl_add_perm (&acle->acle_perms, ACL_WRITE) != 0)
	    goto error;
	if(buf.st_mode & S_IXUSR)
	  if (acl_add_perm (&acle->acle_perms, ACL_EXECUTE) != 0)
	    goto error;

	acle = NULL;
	if (acl_create_entry (&acl, &acle) != 0)
	  goto error;
	if (acl_set_tag_type (acle, ACL_GROUP_OBJ) != 0)
	  goto error;
	if (acl_set_mode (acle, ACL_ABSOLUTE_ENTRY) != 0)
	  goto error;
	acl_clear_perms(&acle->acle_perms);
	if(buf.st_mode & S_IRGRP)
	  if (acl_add_perm (&acle->acle_perms, ACL_READ) != 0)
	    goto error;
	if(buf.st_mode & S_IWGRP)
	  if (acl_add_perm (&acle->acle_perms, ACL_WRITE) != 0)
	    goto error;
	if(buf.st_mode & S_IXGRP)
	  if (acl_add_perm (&acle->acle_perms, ACL_EXECUTE) != 0)
	    goto error;

	acle = NULL;
	if (acl_create_entry (&acl, &acle) != 0)
	  goto error;
	if (acl_set_tag_type (acle, ACL_OTHER_OBJ) != 0)
	  goto error;
	if (acl_set_mode (acle, ACL_ABSOLUTE_ENTRY) != 0)
	  goto error;
	acl_clear_perms(&acle->acle_perms);
	if(buf.st_mode & S_IROTH)
	  if (acl_add_perm (&acle->acle_perms, ACL_READ) != 0)
	    goto error;
	if(buf.st_mode & S_IWOTH)
	  if (acl_add_perm (&acle->acle_perms, ACL_WRITE) != 0)
	    goto error;
	if(buf.st_mode & S_IXOTH)
	  if (acl_add_perm (&acle->acle_perms, ACL_EXECUTE) != 0)
	    goto error;

	return acl;
      error:
	free(acl);
	return NULL;
}
